/*  resident.c
 */

#include <dos.h>
#include <stdio.h>
#include <bios.h>

static union REGS reg;
static struct SREGS seg;
static unsigned mcbseg;
static unsigned dosseg;
static unsigned dosbusy;
static unsigned enddos;
char far *intdta;
static unsigned intsp;
static unsigned intss;
static char far *mydta;
static unsigned myss;
static unsigned stack;
static unsigned ctrl_break;
static unsigned mypsp;
static unsigned intpsp;
static unsigned pids[2];
static int pidctr = 0;
static int pp;
static void interrupt (*oldtimer)(void);
static void interrupt (*old28)(void);
static void interrupt (*oldkb)(void);
void interrupt (*olddisk)(void);
static void interrupt (*oldcrit)(void);
static void interrupt newtimer(void);
static void interrupt new28(void);
static void interrupt newkb(void);
static void interrupt newcrit(int,int,int,int,int,int,int,int,int,int,int,int);
static int resoff = 0;
static int running = 0;
static int popflg = 0;
int diskflag = 0;
int process_done = 0;
unsigned int wait_a_min = 0;
static int kbval;
static int cflag;
extern void interrupt newdisk(int,int,int,int,int,int,int,int,int,int,int,int);
extern void interrupt (*_ZeroDivision)(void);
extern unsigned sizeprogram;
extern unsigned scancode;
extern unsigned keymask;
extern unsigned boot_hour;

/* ------------------------------------------------------------------------ */
static void dores(void);
static void pidaddr(void);
static void resterm(void);
void resident_psp(void);
void interrupted_psp(void);
void popup(void);

/* ------------------------------------------------------------------------ */
/* establish & declare residency
 */
void resinit()
   {
	segread(&seg);
	myss = seg.ss;
	reg.h.ah = 0x34;                    /*  get address of DOS busy flag *
	intdos(&reg, &reg);
	dosseg = _ES;
	dosbusy = reg.x.bx;
	mydta = getdta();               /* get address of resident program's dta */
	pidaddr();                          /* get addresses of PID in DOS */
	                                    /* get original interrupt vectors */
	oldtimer = getvect(0x1c);
	old28 = getvect(0x28);
/***	oldkb = getvect(9);     GMB  leave it here for next time***/
	olddisk = getvect(0x13); 
	                                    /* attach vectors to resident program*/
	setvect(0x1c, newtimer);
/***	setvect(9, newkb);    GMB ***/
	setvect(0x28, new28);
	setvect(0x13, newdisk);  
	                                    /* compute stack pointer  */
	stack = (sizeprogram - (seg.ds - seg.cs)) * 16 - 300;
	                                    /* restore zero divide interrupt vector */
   setvect(0, _ZeroDivision);
	                                    /* terminate and stay resident */
	reg.x.ax = 0x3100;
	reg.x.dx = sizeprogram;
	intdos(&reg, &reg);
   }
#if 0
/* ------------------------------------------------------------------------ */
/*  BIOS disk functions ISR - nice try but messed up formats, see ASM version
 */
static void interrupt newdisk(bp,di,si,ds,es,dx,cx,bx,ax,ip,cs,flgs)
   {
	++diskflag;
	(*olddisk)();
	ax = _AX;		/* for the register returns */
	cx = _CX;
	dx = _DX;
	newcrit(bp,di,si,ds,es,dx,cx,bx,ax,ip,cs,flgs);		/* to get current flags register */
	flgs = cflag;
	--diskflag;
   }
#endif
/* ------------------------------------------------------------------------ */
/*  critical error ISR
 */
static void interrupt newcrit(bp,di,si,ds,es,dx,cx,bx,ax,ip,cs,flgs)
   {
	ax = 0;
	cflag = flgs;	                     /* for newdisk          */
   }
/* ------------------------------------------------------------------------ */
/*  timer ISR
 */
static void interrupt newtimer()
   {
   int hours;
   static char start = NULL;
   extern long far *ticks;

	(*oldtimer)();
   if (!running)
      {
      hours = (int) (((*ticks * 1000L) / 18207L) / 3600L);
      if (wait_a_min > 0 ) --wait_a_min;
      if (start == NULL)
         if(hours != boot_hour)  start = 1;
      if (hours == boot_hour && !resoff && start == 1)
         popflg = 1;
      if (popflg && peekb(dosseg, dosbusy) == 0)
	      if (diskflag == 0 && wait_a_min == 0)
            {
			   outportb(0x20, 0x20);         /* reset timer interrupt */
			   popflg = 0;
			   dores();
		      }
      }
   }

/* ------------------------------------------------------------------------ */
/* alternate keyboard busy loop - if DOS is busy but waiting for akey stroke
 * it should be safe to proceed.
 */
static void interrupt new28()
   {
	(*old28)();
	if (popflg && peekb(dosseg, dosbusy) != 0)
      if (wait_a_min == 0)
         {
		   popflg = 0;
		   dores();
	      }
   }

/* ------------------------------------------------------------------------ */
/*  switch psp context from interrupted to TSR
 */
void resident_psp()
   {
	intpsp = peek(dosseg, *pids);       /* save interrupted program's psp   */
	for (pp = 0; pp < pidctr; pp++)     /* set resident program's psp       */
		poke(dosseg, pids [pp], mypsp);
   }

/* ------------------------------------------------------------------------ */
/*  switch psp context from TSR to interrupted
 */
void interrupted_psp()
   {
	for (pp = 0; pp < pidctr; pp++)     /* reset interrupted program's psp */
		poke(dosseg, pids [pp], intpsp);
   }

/* ------------------------------------------------------------------------ */
/*  switch context to the TSR, execute the resident program, reset to the
 * transient
 */
static void dores()
   {
	running = 1;                        /* so we don't interrup ourself */
	disable();
	intsp = _SP;                        /* swap TSR and transient stacks */
	intss = _SS;
	_SP = stack;
	_SS = myss;
	enable();
	oldcrit = getvect(0x24);            /* redirect critical error     */
	setvect(0x24, newcrit);
	reg.x.ax = 0x3300;		            /* get ctrl break setting      */
	intdos(&reg, &reg);
	ctrl_break = reg.h.dl;
	reg.x.ax = 0x3301;		            /* turn off ctrl break logic   */
	reg.h.dl = 0;
	intdos(&reg, &reg);
	intdta = getdta();		            /* get interrupted dta         */
	setdta(mydta);			               /* set resident dta            */
	resident_psp();			            /* swap psps                   */
	popup();				                  /* execute resident program    */
	interrupted_psp();		            /* reset interrupted psp       */
	setdta(intdta);			            /* reset interrupted dta       */
	setvect(0x24, oldcrit);	            /* reset critical error        */
	reg.x.ax = 0x3301;		            /* reset ctrl break            */
	reg.h.dl = ctrl_break;
	intdos(&reg, &reg);
	disable();				               /* reset interrupted stack     */
	_SP = intsp;
	_SS = intss;
	enable();
	running = 0;
   }

static int avec = 0;
/* -------------------------------------------------------------------------*/
/* test to see if the program is already resident if not, attach to an
 * available interrupt if res parm is set.
 */
unsigned resident(char *signature, void interrupt (*ifunc)(), int res)
   {
	char *sg;
	unsigned df;
	int vec;
    
	segread(&seg);
	df = seg.ds-seg.cs;
	for (vec = 0x60; vec < 0x68; vec++)
      {
		if (getvect(vec) == NULL)
         {
			if (!avec)
				avec = vec;
			continue;
		   }
		for (sg = signature; *sg; sg++)
			if (*sg!=peekb(peek(0,2+vec*4)+df,(unsigned)sg))
				break;
		if (!*sg)
			return vec;
	   }
   if (res)
      {
	   if (avec)
		   setvect(avec, ifunc);
      else return (-1);
      }
	return 0;
   }
/* ------------------------------------------------------------------------ */
/*  find address of PID - at load time use undocumented DOS function 51 to
 *  get Process ID.  Then use equally undocumented # 52 which returns an
 * address of a table of DOS pointers.  The PID is in there somewhere, twice
 * in DOS 2.x.  when the found address matches the returned change it and
 * use 51 again to see if it's the correct one, if not try again.  50 does
 * the set.
 */
static void pidaddr()
   {
	unsigned adr = 0;

	reg.h.ah = 0x51;                    /* get the current pid */
	intdos(&reg, &reg);
	mypsp = reg.x.bx;
	reg.h.ah = 0x52;                    /* find those DOS pointers */
	intdos(&reg, &reg);
	enddos = _ES;
	enddos = peek(enddos, reg.x.bx-2);
	while (pidctr < 2 &&             /* search for matches on the pid in dos */   
			(unsigned)((dosseg<<4) + adr) < (enddos<<4))
      {
		if (peek(dosseg, adr) == mypsp)
         {
			reg.h.ah = 0x50;
			reg.x.bx = mypsp + 1;
			intdos(&reg, &reg);
			if (peek(dosseg, adr) == mypsp+1)
				pids[pidctr++] = adr;
			reg.h.ah = 0x50;              /*  reset the original pid  */
			reg.x.bx = mypsp;
			intdos(&reg, &reg);
   		}
		adr++;
	   }
   }
/* ------------------------------------------------------------------------ */
/*  terminate function , clean all ISR's and release memory
 */
static void resterm()
   {
	void closefiles(void);

	closefiles();	                     /*  close TSR files */
	setvect(0x1c, oldtimer);            /* restore the interrupt vectors */
/***	setvect(9, oldkb);  GMB ***/
	setvect(0x28, old28);
	setvect(0x13, olddisk);
	setvect(avec, (void interrupt (*)()) 0);
	reg.h.ah = 0x52;                    /* get the seg addr of 1st DOS MCB */
	intdos(&reg, &reg);
	mcbseg = _ES;
	mcbseg = peek(mcbseg, reg.x.bx-2);
	segread(&seg);                      /* walk thru mcb chain & release memory */
	while (peekb(mcbseg, 0) == 0x4d)
      {
		if (peek(mcbseg, 1) == mypsp)
         {
			reg.h.ah = 0x49;
			seg.es = mcbseg+1;
			intdosx(&reg, &reg, &seg);
		   }
		mcbseg += peek(mcbseg, 3) + 1;
	   }
   }
/* ------------------------------------------------------------------------ */
/*  terminate the resident program or suspend if we notice someone grabbed
 *  an interrupt, therefore loaded above us.
 */
void terminate()
   {
	if (getvect(0x13) == (void interrupt (*)()) newdisk)
/***		if (getvect(9) == newkb)    GMB  ****/
			if (getvect(0x28) == new28)
				if (getvect(0x1c) == newtimer)
               {
					resterm();
					return;
				   }
	resoff = 1;	                        /* another TSR is above us, suspend */
   }


/* ------------------------------------------------------------------------ */
/*  restart the resident program - un-suspend?
 */
void restart()
   {
	resoff = 0;
   }

/* ------------------------------------------------------------------------ */
/*  put the program on hold, aka suspend
 */
void wait()
   {
	resoff = 1;
   }
