/* ----------------------------------------------------------------------- */
/* TIMEBOOT.c - at BOOT hour TIMEBOOT looks for the target file if specified.
 * If the target is found or not specified TIMEBOOT looks for the alternate
 * batch file.  If found AUTOEXEC.BAT is saved, the alternate is swapped and
 * the PC is warm booted. 
 *
 * This is a TSR, much of the code exists to allow DOS functions to be used
 * in the "background".  Make changes cautiously.
 * Written by Gary Brookman with much code stolen from Al Stevens
 * January 1990.  See the doc file for particulars.
 * If you find any egregious errors please let me know.  I can be reached
 * through Northern Lights BBS.
 */

#define BACKUP   "C:\\BACKUP.ADB"      /* backup Autoexec Dot Bat */
#define BUFSIZE    64                  /* buffer size for copy     */
#define PGMSIZE  7836L                 /* bytes to reserve for TSR */
                                       /* failed at 7788           */    
/* the PGMSIZE is determined by guestimation, look at .map or EXE for
 * a starting point.  Total depends on functions being used.
 */

#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <dir.h>
#include <conio.h>
#include <bios.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <sys\types.h>

/* --------------------------- external variables ------------------------- */
typedef void (far * bootfunc_p)( void );
static union REGS regs;
unsigned long far *ticks;
unsigned sizeprogram = (unsigned) (PGMSIZE/16); 
unsigned boot_hour;
static int process_done;
char restore = 0;
char signature []  = "TB";
char original[]    = "C:\\AUTOEXEC.BAT";
char destination[] = BACKUP;
char alternate[MAXPATH];
char target[MAXPATH];

/* -------------------------- local prototypes ---------------------------- */
unsigned resident(char *, void interrupt (*)(), int);
void resinit(void);
void terminate(void);
void restart(void);
void wait(void);
void resident_psp(void);
void interrupted_psp(void);
void beep(unsigned int);
void closefiles(void);
void popup(void);
void timeboot(void);
int isthere(char *);

/* ----------------------------------------------------------------------- */
main(argc, argv)
char *argv[];
   {
	void interrupt ifunc();
	int ivec;
   if (argv[1][0] == 'i') timeboot();
	if ((ivec = resident(signature, ifunc, 0)) != 0)
      {
		/* ----- TSR is resident ------- */
		if (argc > 1)
         {
			/* ---- there is a command line parameter --- */
			regs.x.ax = 0;
			if (argv[1][0] == 'u') 	      regs.x.ax = 1;
			else if (argv[1][0] == 'r')   regs.x.ax = 2;
			else if (argv[1][0] == 's')   regs.x.ax = 3;
			if (regs.x.ax)
            {
				int86(ivec, &regs, &regs); /* call our interrupt  */
				return;
			   }
		   }
      beep(0);                         /* can't go resident again */
      exit(1);
	   }
 	else                                /* initial load             */
      {
      if (argc < 3)
         {
         beep(0);
         exit (1);
         }
      if ((boot_hour = atoi(argv[1])) > 23) /* force midnight for non # */
         {
         beep(0);
         exit (1);
         }
/** probably not good practice but TURBOC does trim trailing white space **/
      if (argv[2][0] == 'r' && argv[2][1] == '\0')
         {
         strcpy(alternate,BACKUP);
         restore = 1;
         }
      else
         strcpy(alternate,argv[2]);
      if (! isthere(alternate))
         {
         beep(0);
         beep(9);
         exit (1);
         }
      if (argc > 3) strcpy(target,argv[3]);
      else process_done = 1;
	if ((ivec = resident(signature, ifunc, 1)) != 0)
         {
         beep(0);
         exit (1);
         }
/*  ticks points to the BIOS area used to store the # of ticks since 00:00 */
	   ticks = (((unsigned long)(0x0040) << 16) | (unsigned)(0x006c));
      beep(5);                         
		resinit();                       /* t and sr                */

	}
}

/* ------------------------------------------------------------------------ */
/* our own interrupt, # based on what's available
 */
void interrupt ifunc(bp,di,si,ds,es,dx,cx,bx,ax)
   {
	if (ax == 1)			/* "unload" */
		terminate();
	else if (ax == 2)		/* "restart" */
		restart();
	else if (ax == 3)		/* "suspend" */
		wait();
   }
/* ------------------------------------------------------------------------ */
/* not used in this incarnation
 */
void closefiles()
   {
   }

/* ------------------------------------------------------------------------ */
/* this function is called by the TSR portion when boot_hour is reached.  if
 * a target file was named we wait until it's there and closed.  Then, swap
 * the batch files and boot.
 */
void popup()
   {
   extern unsigned int wait_a_min;

   if (!process_done)
      if (! (process_done = isthere(target)) ) 
         {
         beep(9);
         wait_a_min = 1092;            /* around 1 minute */
         return;
         }
   if (!restore)
      if (copy(original, destination) < 0)
         {
         beep(0);
         wait_a_min = 1092;               
         return;
         }
   if (copy(alternate, original) < 0)
      {
      beep(0);
      wait_a_min = 1092;               
      return;
      }
   
	timeboot();					            /* call the boot function */

   }
      
/* ----------------------------------------------------------------------- */
void beep(unsigned int tone)
   {
   unsigned int  j;

   if (tone == 0) ++tone;
   sound(tone * 128); 
   for (j=0; j < 40000; j++) ;         /*  for 1/2 second. */
   nosound();                          /* Turn off speaker  */
   return;
   }

/* ----------------------------------------------------------------------- */
void timeboot(void)
   {        
   int far    *boottype;                 /* pointer to boot type    */
   bootfunc_p  boot;                     /* pointer to boot routine */
 
   boottype  = (int far *)0x0472;        /* point to boot type      */
   *boottype = 0x1234;                   /* set boot type           */
   boot      = (bootfunc_p)0xFFFF0000L;  /* point to boot address   */
   boot();                               /* go to it                */
   }
/* ----------------------------------------------------------------------- */
int copy(char *infile, char *outfile)
   {
   int in, out, n;
   char buf[BUFSIZE];

   if ((in = open(infile, 0)) == -1) return (-1);
   if ((out = open(outfile, O_CREAT|O_WRONLY|O_TRUNC,S_IWRITE)) == -1)
      return (-1);
   while ((n = read(in, buf, BUFSIZE-1)) > 0)
      if (write(out,buf,n) != n) return (-1);
   close(in);
   close(out);
   return (1);
   }
/* ------------------------------------------------------------------------ */
/* check for existance of file, and that it's not zero length.  Zero length
 * is assumed to mean the file is still open.  Return 1 if it's there, 0
 * otherwise.
 */
int isthere(char *filename)
   {
   char far *dtap;

   if (access(filename,02) == 0)       /* does file exist   */
      {
      regs.h.ah = 0x4e;
      regs.h.ch = 0x0;
      regs.h.cl = 0x0;
      regs.x.dx = (int) filename;
      intdos(&regs,&regs);
      dtap = getdta();
      if ((unsigned long) *((long far *) (dtap+0x1a)) > 0L)  return (1); 
      }                                /* file not there    */
   return (0);
   }	
